from typing import Any, Optional, Tuple
import hydra
from omegaconf import DictConfig
from copy import deepcopy

import torch
import torch.nn.functional as F

import pytorch_lightning as pl

from torch.utils.data.dataloader import DataLoader

from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.samplers import BatchSamplerDefault
from fuse.dl.models.model_wrapper import ModelWrapSeqToDict
from fuse.dl.lightning.pl_module import LightningModuleDefault
from fuse.dl.models.heads import Head1D
from fuse.eval.metrics.classification.metrics_classification_common import MetricAUCROC
from fuse.dl.losses import LossDefault
from fuse.utils import NDict

from fuse_examples.multimodality.ehr_transformer.model import (
    Embed,
    TransformerEncoder,
    Bert,
    BertConfig,
)
from fuse_examples.multimodality.ehr_transformer.dataset import PhysioNetCinC


def filter_gender_label_unknown(batch_dict: NDict) -> NDict:
    """
    Ignore unlabeled (gender) when computing gender classification loss
    """
    # filter out samples
    keep_indices = batch_dict["Gender"] != -1
    batch_dict = batch_dict.indices(
        keep_indices
    )  # this will keep only a subset of elements per each key in the dictionary (where applicable)
    return batch_dict


def data(
    dataset_cfg: dict,
    target_key: str,
    batch_size: int,
    data_loader_train: dict,
    data_loader_valid: dict,
) -> Tuple[Any, DataLoader, DataLoader]:
    """
    return token to index mapper and train and validation dataloaders for MIMICC II
    :param dataset_cfg: PhysioNetCinC.dataset arguments
    :param target_key: will be used to balance the training dataset
    :param data_loader_train: arguments for train dataloader
    :param data_loader_train: arguments for validation dataloader
    """

    token2idx, ds_train, ds_valid, _ = PhysioNetCinC.dataset(**dataset_cfg)

    dl_train = DataLoader(
        ds_train,
        collate_fn=CollateDefault(
            keep_keys=[
                "data.sample_id",
                "Target",
                "Indexes",
                "Gender",
                "NextVisitLabels",
            ]
        ),
        batch_sampler=BatchSamplerDefault(
            ds_train,
            balanced_class_name=target_key,
            batch_size=batch_size,
            mode="approx",
            num_batches=500,
        ),
        **data_loader_train,
    )
    dl_valid = DataLoader(
        ds_valid,
        collate_fn=CollateDefault(
            keep_keys=[
                "data.sample_id",
                "Target",
                "Indexes",
                "Gender",
                "NextVisitLabels",
            ]
        ),
        **data_loader_valid,
    )

    return token2idx, dl_train, dl_valid


def model(
    embed: dict,
    classifier_head: dict,
    z_dim: int,
    encoder_type: str,
    transformer_encoder: dict,
    bert_config_kwargs: dict,
    vocab_size: int,
    aux_gender_classification: bool,
    classifier_gender_head: dict,
    aux_next_vis_classification: bool,
    classifier_next_vis_head: dict,
) -> torch.nn.Module:
    """
    Create transformer based model with 3 classification heads
    :param embed: arguments for Embed constructor
    :param classifier_head: arguments for Head1D for main classification task
    :param z_dim: the output size of the transformer model (cls token)
    :param encoder_type: either "transformer" or "bert"
    :param transformer_encoder: arguments for TransformerEncoder - used when encoder_type is "transformer"
    :param bert_config_kwargs: arguments for BertConfig - used with encoder_type is "bert"
    :param vocab_size: vocabulary size
    :param  aux_gender_classification: enable gender auxiliary classification head
    :param classifier_gender_head: arguments for Head1D for gender classification task
    :param  aux_next_vis_classification: enable next visit auxiliary classification head
    :param classifier_next_vis_head: arguments for Head1D for next visit classification task
    """
    if encoder_type == "transformer":
        encoder_model = ModelWrapSeqToDict(
            model=TransformerEncoder(**transformer_encoder),
            model_inputs=["model.embedding"],
            model_outputs=["model.z", None],
        )
    elif encoder_type == "bert":
        bert_config = BertConfig(
            vocab_size_or_config_json_file=vocab_size, **bert_config_kwargs
        )

        encoder_model = ModelWrapSeqToDict(
            model=Bert(config=bert_config),
            model_inputs=["model.embedding"],
            model_outputs=["model.z"],
        )
    else:
        raise Exception(f"Error: unknown encoder_type {encoder_type}")

    models_sequence = [
        Embed(key_in="Indexes", key_out="model.embedding", n_vocab=vocab_size, **embed),
        encoder_model,
        Head1D(head_name="cls", conv_inputs=[("model.z", z_dim)], **classifier_head),
    ]

    # append auxiliary head for gender
    if aux_gender_classification:
        models_sequence.append(
            Head1D(
                head_name="gender",
                conv_inputs=[("model.z", z_dim)],
                **classifier_gender_head,
            )
        )

    if aux_next_vis_classification:
        models_sequence.append(
            Head1D(
                head_name="next_vis",
                conv_inputs=[("model.z", z_dim)],
                num_outputs=vocab_size,
                **classifier_next_vis_head,
            )
        )

    model = torch.nn.Sequential(*models_sequence)
    return model


def train(
    model: torch.nn.Module,
    dl_train: DataLoader,
    dl_valid: DataLoader,
    model_dir: str,
    opt: callable,
    trainer_kwargs: dict,
    target_key: str,
    target_loss_weight: float,
    aux_gender_classification: bool,
    gender_loss_weight: float,
    aux_next_vis_classification: bool,
    next_vis_loss_weight: float,
    lr_scheduler: callable = None,
    track_clearml: Optional[dict] = None,
) -> None:
    """
    Run training process
    :param model: the model to train
    :param dl_train: train set dataloader
    :param dl_valid: validation set dataloader
    :param model_dir: path to store the training process outputs
    :param opt: callable that give model parameters will return torch optimizer
    :param trainer_kwargs: parameters to pl.Trainer
    :param target_key: key that points to main task labels
    :param target_loss_weight: weight for main task loss
    :param aux_gender_classification: enable gender auxiliary classification head
    :param gender_loss_weight: weight for gender classification task loss
    :param aux_next_vis_classification: enable next visit auxiliary classification head
    :param next_vis_loss_weight: weight for next visit classification task loss
    :param lr_scheduler: callable that given optimizer returns torch lr scheduler
    :param track_clearml: optional - to track with clearml provide arguments to start_clearml_logger()
    """

    if track_clearml is not None:
        from fuse.dl.lightning.pl_funcs import start_clearml_logger

        start_clearml_logger(**track_clearml)

    #  Loss
    losses = {
        "ce": LossDefault(
            pred="model.logits.cls",
            target=target_key,
            callable=F.cross_entropy,
            weight=target_loss_weight,
        ),
    }

    # Metrics
    train_metrics = {
        "auc": MetricAUCROC(pred="model.output.cls", target=target_key),
    }

    # auxiliary gender loss and metric
    if aux_gender_classification:
        losses["gender_ce"] = LossDefault(
            pred="model.logits.gender",
            target="Gender",
            callable=F.cross_entropy,
            preprocess_func=filter_gender_label_unknown,
            weight=gender_loss_weight,
        )

        train_metrics["gender_auc"] = MetricAUCROC(
            pred="model.output.gender",
            target="Gender",
            batch_pre_collect_process_func=filter_gender_label_unknown,
        )

    # auxiliary gender loss and metric
    if aux_next_vis_classification:
        losses["next_vis_ce"] = LossDefault(
            pred="model.logits.next_vis",
            target="NextVisitLabels",
            callable=torch.nn.BCEWithLogitsLoss(),  # multi binary labels loss
            weight=next_vis_loss_weight,
        )

    validation_metrics = deepcopy(train_metrics)
    # either a dict with arguments to pass to ModelCheckpoint or list dicts for multiple ModelCheckpoint callbacks (to monitor and save checkpoints for more then one metric).
    best_epoch_source = dict(
        monitor="validation.metrics.auc",
        mode="max",
    )

    # optimizer and lr sch - see pl.LightningModule.configure_optimizers return value for all options
    optimizer = opt(params=model.parameters())
    optimizers_and_lr_schs = dict(optimizer=optimizer)

    if lr_scheduler is not None:
        optimizers_and_lr_schs["lr_scheduler"] = lr_scheduler(optimizer)
        if isinstance(
            optimizers_and_lr_schs["lr_scheduler"],
            torch.optim.lr_scheduler.ReduceLROnPlateau,
        ):
            optimizers_and_lr_schs["monitor"] = "validation.losses.total_loss"

    #  Train

    # create instance of PL module - FuseMedML generic version
    pl_module = LightningModuleDefault(
        model_dir=model_dir,
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=validation_metrics,
        best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
    )

    # create lightning trainer.
    pl_trainer = pl.Trainer(**trainer_kwargs)

    # train
    pl_trainer.fit(pl_module, dl_train, dl_valid, ckpt_path=None)


@hydra.main(config_path=".", config_name="config")
def main(cfg: DictConfig) -> None:
    print(str(cfg))

    cfg = hydra.utils.instantiate(cfg)

    # get data
    token2idx, dl_train, dl_valid = data(**cfg.data)

    # create model
    nn_model = model(vocab_size=len(token2idx), **cfg.model)

    # train the model
    train(model=nn_model, dl_train=dl_train, dl_valid=dl_valid, **cfg.train)


if __name__ == "__main__":
    """
    See README for instructions
    """
    main()



from typing import Dict
import torch
from torch import nn
from fuse.utils.ndict import NDict


class LossWrapToDict(torch.nn.Module):
    """
    Wraps a torch loss function to support a batch dict.
    See usage_example() below for additional details.
    """

    def __init__(
        self,
        *,
        loss_module: torch.nn.Module,
        loss_arg_to_batch_key: Dict[str, str],
        weight: float = 1.0,
    ) -> None:
        """
        :param loss_module: the loss module to wrap
        :param loss_arg_to_batch_key: each key is an argument in the forward function,
                                      each value is the corresponding key in the batch_dict
        """
        super().__init__()
        self._loss_module = loss_module
        self.loss_arg_to_batch_key = loss_arg_to_batch_key
        self._weight = weight

    def forward(self, batch_dict: NDict) -> torch.Tensor:
        # collect arguments for loss module
        loss_kwargs = {
            arg: batch_dict[batch_key]
            for arg, batch_key in self.loss_arg_to_batch_key.items()
        }
        # run loss function
        loss = self._loss_module(**loss_kwargs)
        return self._weight * loss


def usage_example() -> torch.Tensor:
    batch_size = 100
    num_classes = 10
    # creating an example of a batch_dict
    logits = torch.randn(size=(batch_size, num_classes))
    gt = torch.randint(high=num_classes, size=(batch_size,))
    batch_dict = NDict({"data.gt": gt, "model.logits": logits})

    # creating the loss
    loss = nn.CrossEntropyLoss()
    # the CrossEntropyLoss forward function has two arguments: input, target
    # this maps every forward argument to the relevant key in the batch_dict
    loss_arg_to_batch_key = {"input": "model.logits", "target": "data.gt"}
    # wrapping the loss
    wrapped_loss = LossWrapToDict(
        loss_module=loss, loss_arg_to_batch_key=loss_arg_to_batch_key
    )
    # running the wrapped loss with the batch_dict
    loss_result = wrapped_loss(batch_dict)
    print(f"the output loss is: {loss_result}")
    return loss_result


if __name__ == "__main__":
    usage_example()



from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuseimg.data.ops.color import OpNormalizeAgainstSelf

from fuseimg.data.ops.aug.geometry import (
    OpAugAffine2D,
    OpAugSqueeze3Dto2D,
    OpAugUnsqueeze3DFrom2D,
)
from fuse.data import PipelineDefault, OpToTensor, OpRepeat
from fuse.data.ops.ops_common import OpLambda, OpLookup, OpToOneHot

from fuse.data.ops.ops_aug_common import OpRandApply, OpSampleAndRepeat
from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.ops.ops_cast import OpToNumpy
from fuse.data.ops.op_base import OpBase
from fuse.utils import NDict
from functools import partial
import nibabel as nib
from typing import Hashable, Optional, Sequence
import torch
import pandas as pd
import numpy as np
import skimage
import os
from fuse.data.utils.sample import get_sample_id
from medpy.io import load

from fuse.utils.rand.param_sampler import Uniform, RandInt


class OpPICAISampleIDDecode(OpBase):
    """
    decodes sample id into image and segmentation filename
    """

    def __call__(self, sample_dict: NDict) -> NDict:
        """ """
        sid = get_sample_id(sample_dict)

        img_filename_key = "data.input.img_path"
        sample_dict[img_filename_key] = sid

        return sample_dict


class OpLoadPICAIImage(OpBase):
    """
    Loads a medical image
    """

    def __init__(
        self, dir_path: str, seuqences: Sequence[str] = ["_t2w"], **kwargs: dict
    ):
        super().__init__(**kwargs)
        self._dir_path = dir_path
        self._sequences = seuqences

    def __call__(
        self,
        sample_dict: NDict,
        key_in: str,
        key_out: str,
    ) -> None:
        """
        :param key_in: the key name in sample_dict that holds the filename
        :param key_out: the key name in sample_dict that holds the image
        :param key_metadata_out : the key to hold metadata dictionary
        """
        try:
            for seq in self._sequences:
                img_filename = os.path.join(
                    self._dir_path,
                    sample_dict[key_in].split("_")[0],
                    sample_dict[key_in] + seq + ".mha",
                )

                image_data, image_header = load(img_filename)
                sample_dict[key_out + seq] = image_data
            return sample_dict
        except:
            print(
                "sample id",
                sample_dict[key_in].split("_")[0],
                "has missing sequence",
                seq,
                "file name does not exist",
                img_filename,
            )
            return None


# loads cancer segmentation - not in use for now
class OpLoadPICAISegmentation(OpBase):
    """
    Loads a medical image
    """

    def __init__(self, data_dir: str, dir_path: str, **kwargs: dict):
        super().__init__(**kwargs)
        self._dir_path = dir_path
        self._data_dir = data_dir

    def __call__(
        self,
        sample_dict: NDict,
        key_in: str,
        key_out: str,
        gt: str,
    ) -> None:
        """
        :param key_in: the key name in sample_dict that holds the filename
        :param key_out: the key name in sample_dict that holds the image
        :param key_metadata_out : the key to hold metadata dictionary
        """
        # for seq in self._sequences:
        seg_filename = os.path.join(self._dir_path, sample_dict[key_in] + ".nii.gz")
        if os.path.exists(seg_filename):
            my_img = nib.load(seg_filename)
            nii_data = my_img.get_fdata()
        else:
            img_filename = os.path.join(
                self._data_dir,
                sample_dict[key_in].split("_")[0],
                sample_dict[key_in] + "_t2w.mha",
            )
            image_data, image_header = load(img_filename)
            nii_data = np.zeros(image_data.shape)
        unique_labels = np.unique(nii_data)
        if sample_dict[gt] not in unique_labels:
            print(
                sample_dict[key_in],
                "ISUP is",
                sample_dict[gt],
                "but unique labels are",
                unique_labels,
            )
            return None
        # TODO - generate different map according to different ISUP in cancer segmentation?
        # else:
        #     output = np.zeros([5]+image_data.shape)
        #     for label in unique_labels:
        #         if label != 0:
        #             output[label-1] = nii_data[nii_data == label]
        sample_dict[key_out] = nii_data
        return sample_dict


class OpLoadPICAISegmentationWholeGland(OpBase):
    """
    Loads a medical image
    """

    def __init__(self, data_dir: str, dir_path: str, **kwargs: dict):
        super().__init__(**kwargs)
        self._dir_path = dir_path
        self._data_dir = data_dir

    def __call__(
        self,
        sample_dict: NDict,
        key_in: str,
        key_out: str,
        gt: str,
    ) -> None:
        """
        :param key_in: the key name in sample_dict that holds the filename
        :param key_out: the key name in sample_dict that holds the image
        :param key_metadata_out : the key to hold metadata dictionary
        """
        # for seq in self._sequences:
        seg_filename = os.path.join(self._dir_path, sample_dict[key_in] + ".nii.gz")
        if os.path.exists(seg_filename):
            my_img = nib.load(seg_filename)
            nii_data = my_img.get_fdata()
        else:
            print(seg_filename, "missing segmentation file")
            return None
        sample_dict[key_out] = nii_data
        return sample_dict


class PICAI:
    """ """

    @staticmethod
    def static_pipeline(
        data_source: pd.DataFrame,
        data_dir: str,
        seg_dir: str,
        target: str,
        repeat_images: Sequence[NDict],
    ) -> PipelineDefault:
        """
        Get suggested static pipeline (which will be cached), typically loading the data plus design choices that we won't experiment with.
        :param data_path: path to original kits21 data (can be downloaded by KITS21.download())
        """
        repeat_images_with_seg = repeat_images + [(dict(key="data.gt.seg"))]
        bool_map = {"NO": 0, "YES": 1}
        static_pipeline = PipelineDefault(
            "cmmd_static",
            [
                # decoding sample ID
                (
                    OpPICAISampleIDDecode(),
                    dict(),
                ),  # will save image and seg path to "data.input.img_path"
                (
                    OpLoadPICAIImage(data_dir),
                    dict(key_in="data.input.img_path", key_out="data.input.img"),
                ),
                (
                    OpReadDataframe(
                        data_source,
                        key_column="index",
                        key_name="data.input.img_path",
                        #'psa','psad','prostate_volume','histopath_type','lesion_GS','lesion_ISUP','case_ISUP'
                        columns_to_extract=[
                            "index",
                            "patient_id",
                            "study_id",
                            "mri_date",
                            "patient_age",
                            "case_ISUP",
                            "case_csPCa",
                        ],
                        rename_columns=dict(
                            patient_id="data.patientID",
                            case_csPCa="data.gt.classification",
                            case_ISUP="data.gt.subtype",
                        ),
                    ),
                    dict(),
                ),
                (
                    OpLookup(bool_map),
                    dict(
                        key_in="data.gt.classification",
                        key_out="data.gt.classification",
                    ),
                ),
                (
                    OpToOneHot(len(bool_map)),
                    dict(
                        key_in="data.gt.classification",
                        key_out="data.gt.classification_one_hot",
                    ),
                ),
                (
                    OpLoadPICAISegmentationWholeGland(data_dir, seg_dir),
                    dict(
                        key_in="data.input.img_path",
                        key_out="data.gt.seg",
                        gt="data.gt.subtype",
                    ),
                ),
                (
                    OpRepeat(
                        OpLambda(partial(np.transpose, axes=[2, 0, 1])),
                        kwargs_per_step_to_add=repeat_images_with_seg,
                    ),
                    dict(),
                ),
                (
                    OpRepeat(
                        (
                            OpLambda(
                                partial(
                                    skimage.transform.resize,
                                    output_shape=(32, 256, 256),
                                    mode="reflect",
                                    anti_aliasing=True,
                                    preserve_range=True,
                                )
                            )
                        ),
                        kwargs_per_step_to_add=repeat_images_with_seg,
                    ),
                    {},
                ),
                # (OpNormalizeAgainstSelf(), dict(key="data.input.img_t2w")),
                (
                    OpRepeat(
                        (OpNormalizeAgainstSelf()), kwargs_per_step_to_add=repeat_images
                    ),
                    dict(),
                ),
                (
                    OpRepeat((OpToNumpy()), kwargs_per_step_to_add=repeat_images),
                    dict(dtype=np.float32),
                ),
                # (OpResizeAndPad2D(), dict(key="data.input.img", resize_to=(2200, 1200), padding=(60, 60))),
            ],
        )
        return static_pipeline

    @staticmethod
    def dynamic_pipeline(
        repeat_images: Sequence[NDict],
        train: bool = False,
        aug_params: NDict = None,
    ) -> PipelineDefault:
        """
        Get suggested dynamic pipeline. including pre-processing that might be modified and augmentation operations.
        :param train : True iff we request dataset for train purpouse
        """
        repeat_images_with_seg = repeat_images + [dict(key="data.gt.seg")]
        ops = []
        ops += [
            (
                OpRepeat((OpToTensor()), kwargs_per_step_to_add=repeat_images),
                dict(dtype=torch.float32),
            ),
            (OpToTensor(), dict(key="data.gt.seg", dtype=torch.int32)),
            (
                OpRepeat(
                    (OpLambda(partial(torch.unsqueeze, dim=0))),
                    kwargs_per_step_to_add=repeat_images_with_seg,
                ),
                {},
            ),
        ]
        if train:
            ops += [
                # affine augmentation - will apply the same affine transformation on each slice
                (
                    OpRepeat(
                        (OpAugSqueeze3Dto2D()),
                        kwargs_per_step_to_add=repeat_images_with_seg,
                    ),
                    dict(axis_squeeze=1),
                ),
                (
                    OpRandApply(
                        OpSampleAndRepeat(
                            OpAugAffine2D(),
                            kwargs_per_step_to_add=repeat_images_with_seg,
                        ),
                        aug_params["apply_aug_prob"],
                    ),
                    dict(
                        rotate=Uniform(*aug_params["rotate"]),
                        scale=Uniform(*aug_params["scale"]),
                        flip=(aug_params["flip"], aug_params["flip"]),
                        translate=(
                            RandInt(*aug_params["translate"]),
                            RandInt(*aug_params["translate"]),
                        ),
                    ),
                ),
                (
                    OpRepeat(
                        OpAugUnsqueeze3DFrom2D(),
                        kwargs_per_step_to_add=repeat_images_with_seg,
                    ),
                    dict(axis_squeeze=1, channels=1),
                ),
            ]
        dynamic_pipeline = PipelineDefault("picai_dynamic", ops)
        return dynamic_pipeline

    @staticmethod
    def dataset(
        paths: NDict,
        cfg: NDict,
        reset_cache: bool = True,
        sample_ids: Optional[Sequence[Hashable]] = None,
        train: bool = False,
        run_sample: int = 0,
    ) -> DatasetDefault:
        """
        Creates Fuse Dataset single object (either for training, validation and test or user defined set)
        :param paths                        paths dictionary for dataset files
        :param cfg                          dict cfg for training phase
        :param reset_cache:                 Optional,specifies if we want to clear the cache first
        :param sample_ids:                  dataset including the specified sample_ids or None for all the samples.
        :param train:                       True if used for training  - adds augmentation operations to the pipeline
        :param run_sample:                  if > 0 it samples from all the samples #run_sample examples ( used for testing), if =0 then it takes all samples
        :return: DatasetDefault object
        """

        input_source_gt = pd.read_csv(paths["clinical_file"])
        input_source_gt["index"] = (
            input_source_gt["patient_id"].astype(str)
            + "_"
            + input_source_gt["study_id"].astype(str)
        )
        all_sample_ids = input_source_gt["index"].to_list()

        if sample_ids is None:
            sample_ids = all_sample_ids
        if run_sample > 0:
            sample_ids = sample_ids[:run_sample]
        repeat_images = [
            dict(key="data.input.img" + seq) for seq in cfg["series_config"]
        ]
        static_pipeline = PICAI.static_pipeline(
            input_source_gt,
            paths["data_dir"],
            paths["seg_dir"],
            cfg["target"],
            repeat_images,
        )
        if train:
            dynamic_pipeline = PICAI.dynamic_pipeline(
                train=train, repeat_images=repeat_images, aug_params=cfg["aug_params"]
            )
        else:
            dynamic_pipeline = PICAI.dynamic_pipeline(
                train=train, repeat_images=repeat_images
            )

        cacher = SamplesCacher(
            "cache_ver",
            static_pipeline,
            cache_dirs=[paths["cache_dir"]],
            restart_cache=reset_cache,
            audit_first_sample=False,
            audit_rate=None,
            workers=cfg["num_workers"],
        )

        my_dataset = DatasetDefault(
            sample_ids=sample_ids,
            static_pipeline=static_pipeline,
            dynamic_pipeline=dynamic_pipeline,
            cacher=cacher,
        )

        my_dataset.create()
        return my_dataset



from typing import List, Dict, Callable, Any  # , _UnionGenericAlias


import inspect

# https://stackoverflow.com/questions/147816/preserving-signatures-of-decorated-functions
# https://stackoverflow.com/questions/5929107/decorators-with-parameters
# https://stackoverflow.com/questions/2366713/can-a-decorator-of-an-instance-method-access-the-class


"""
This is a very handy class that allows enforcing an API
Unlike programming languages like C++, python does not have a built in way to enforce adherance to API when inheriting from a base class
By using the code found in this script you can require the inheriting class to implement certain method and also control which arguments of the method signature is required.
an actively used example can be seen in InsilicoLabModelBase() in insilico-lab-validation repo.
For illustration, you can create a base class FruitBase (you can see related unit test in `fuse-med-ml/fuse/utils/tests/test_interface_validator.py`)


```python
from fuse.utils.interface_validator import InterfaceValidator, validate_signature

class FruitBase(InterfaceValidator):
    def __init__(self, *, wow='wow'):
        super().__init__()
        self.wow = wow

    @validate_signature(args=['color'])  #only argument `color` will be validated to both exist and have the same type hint in the inheriting class
    def draw(self, *, color:str = 'red', style:int=12, **kwargs):
        print(f'FruitBase.draw: {color=} {style=}')

    @validate_signature(args=['size', 'material']) #both args `size` and `material` will be validated
    def print_3d(self, *, size:float = 2.7, material:str = 'iron'):
        print(f'FruitBase.print_3d {size=} {material=}')
```

now, when you use a class that inherits from it, like:

```python
class Banana(FruitBase):
    #...
```

the API will be enforced and exceptions will be raised if requirements are not met.


#TODO:
    1. allow validating return type hint
    2. allow using @validate_signature without using any args, and that would mean that the exact same full signature is required
    3. allow control over whether type hint is validated or not (is this needed?)
"""


def validate_signature(*, args: List[str] = None) -> Callable:
    def wrapper(function: Callable) -> Any:
        return validate_signature_impl(function, args=args)

    return wrapper


class validate_signature_impl(object):
    def __init__(self, fn: Callable, args: List[str] = None) -> None:
        self.fn = fn
        self._args_to_validate = args

    def __set_name__(self, owner: Any, name: str) -> None:
        # do something with owner, i.e.
        curr_func = self.fn
        print(f"decorating {curr_func} and using {owner}")
        if not hasattr(owner, "_validate_methods"):
            owner._validate_methods = {}
        owner._validate_methods[name] = {}

        signature_info = inspect.getfullargspec(self.fn)

        for curr_arg in self._args_to_validate:
            owner._validate_methods[name][curr_arg] = signature_info.annotations[
                curr_arg
            ]

        # then replace ourself with the original method
        setattr(owner, name, curr_func)


class InterfaceValidator:
    def __init__(self, verbose: bool = False):
        super().__init__()
        self.__dict__["_properly_initialized_InterfaceValidator"] = True
        self._verbose_signature_validation = verbose
        if not hasattr(self, "_validate_methods"):
            raise Exception(
                "InterfaceValidator: error - seems like you did not requested signature validation for any method. To use that you need to use @interface_validator.validate_signature on at least one method"
            )

        for method_name, curr_validate_args in self._validate_methods.items():
            curr_bases = self.__class__.__bases__
            if len(curr_bases) > 1:
                raise Exception("currently not supporting more than one base")
            if len(curr_bases) == 0:
                raise Exception("expected at least one base")

            if getattr(type(self), method_name) == getattr(curr_bases[0], method_name):
                raise Exception(
                    f'a class that inherited from InterfaceValidator did not override interface method "{method_name}" !'
                )

            curr_method = getattr(self, method_name)
            if self._verbose_signature_validation:
                print(
                    f'found method "{curr_method}" with requested validation of signature for args {curr_validate_args}'
                )
            self._validate_signature(method_name, curr_validate_args)

    def _validate_signature(
        self, method_name: str, validate_args: Dict[str, Any]
    ) -> None:
        """
        raises an exception if a signature mismatch is found
        """
        curr_method = getattr(self, method_name)
        actual_method = curr_method

        func_details = inspect.getfullargspec(actual_method)

        pos_args = [x for x in func_details.args if x != "self"]
        if len(pos_args) > 0:
            raise Exception(
                f'found positional args in the function of {actual_method} - this is not supported. Positional args found = {pos_args}. Tip: you can put * after the "self" argument to prevent positional args. For example: def foo(self, *, arg1:int, arg2:float=2.3) '
            )

        for arg_name, arg_type_hint in validate_args.items():
            if arg_name not in func_details.kwonlyargs:
                raise Exception(
                    f'error in class {self.__class__.__name__} method {method_name} - could not find expected kwarg "{arg_name}"'
                )

            if arg_name not in func_details.annotations:
                raise Exception(
                    f"type annotation is required in class {self.__class__.__name__} in method {method_name} - arg {arg_name} "
                )

            actual_type_hint = func_details.annotations[arg_name]
            if actual_type_hint != arg_type_hint:
                raise Exception(
                    f"error in class {self.__class__.__name__} in method {method_name} - arg {arg_name} is expected to have type hint = {arg_type_hint} but got {actual_type_hint} instead"
                )
            if self._verbose_signature_validation:
                print(
                    f"in class {self.__class__.__name__} method {method_name} arg {arg_name} validated ok."
                )

    def __setattr__(self, name: str, value: Any) -> None:
        if not hasattr(self, "_properly_initialized_InterfaceValidator"):
            raise Exception(
                "When you inherit from InterfaceValidator you must call super().__init__() in your constructor as the first thing you do."
            )
        super().__setattr__(name, value)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import shutil
import unittest
import tempfile
import os

from fuse.utils.multiprocessing.run_multiprocessed import run_in_subprocess
from omegaconf import DictConfig
import yaml

from fuse_examples.multimodality.ehr_transformer.main_train import main as main_train


if "CINC_TEST_DATA_PATH" in os.environ:
    TEST_PARAMS = {
        "CONFIG_PATH": os.path.dirname(__file__)
        + "/../multimodality/ehr_transformer/config.yaml",
        "TEST_DATA_DIR": os.environ["CINC_TEST_DATA_PATH"],
        "TEST_DATA_PICKL": None,
        "NAME": "ehr_transformer_test",
        "BATCH_SIZE": 30,
        "NUM_EPOCHS": 2,
    }


def init_test_environment(root: str) -> DictConfig:
    with open(TEST_PARAMS["CONFIG_PATH"], "r") as file:
        cfg = yaml.safe_load(file)

    # update confifuration relevant to the unitest TEST DATA folder
    cfg["root"] = root
    cfg["name"] = TEST_PARAMS["NAME"]
    cfg["data"]["batch_size"] = TEST_PARAMS["BATCH_SIZE"]
    cfg["data"]["dataset_cfg"]["raw_data_path"] = TEST_PARAMS["TEST_DATA_DIR"]
    cfg["data"]["dataset_cfg"]["raw_data_pkl"] = TEST_PARAMS["TEST_DATA_PICKL"]
    cfg["train"]["trainer_kwargs"]["max_epochs"] = TEST_PARAMS["NUM_EPOCHS"]

    return cfg


def run_ehr_transformer(cfg: DictConfig) -> None:
    try:
        main_train(cfg)

    except:
        raise Exception("Training EHR Transformer Failed.")


@unittest.skipIf(
    "CINC_TEST_DATA_PATH" not in os.environ,
    "define environment variable 'CINC_TEST_DATA_PATH' to run this test",
)
class EHRTransformerTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.root = tempfile.mkdtemp()
        self.cfg = init_test_environment(self.root)

    def test_template(self) -> None:
        run_in_subprocess(run_ehr_transformer, self.cfg)

    def tearDown(self) -> None:
        shutil.rmtree(self.root)


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest
import os
import tempfile
import shutil
from fuse.utils.multiprocessing.run_multiprocessed import run_in_subprocess

from fuse_examples.imaging.classification.isic.isic_runner import (
    TRAIN_COMMON_PARAMS,
    INFER_COMMON_PARAMS,
    EVAL_COMMON_PARAMS,
    run_train,
    run_infer,
    run_eval,
)
from fuse_examples.imaging.classification.isic.golden_members import FULL_GOLDEN_MEMBERS

import fuse.utils.gpu as GPU
from fuse.utils.rand.seed import Seed


def run_isic(
    root: str,
    train_common_params: dict,
    infer_common_params: dict,
    eval_common_params: dict,
) -> None:
    model_dir = os.path.join(root, "model_dir")
    paths = {
        "model_dir": model_dir,
        "data_dir": os.environ["ISIC19_DATA_PATH"]
        if "ISIC19_DATA_PATH" in os.environ
        else os.path.join(root, "isic/data_dir"),
        "cache_dir": os.path.join(root, "cache_dir"),
        "inference_dir": os.path.join(model_dir, "infer_dir"),
        "eval_dir": os.path.join(model_dir, "eval_dir"),
        "data_split_filename": os.path.join(root, "split.pkl"),
    }

    dd = paths["data_dir"]
    print(f"data_dir={dd}")

    # Must use GPU due a long running time
    GPU.choose_and_enable_multiple_gpus(1, use_cpu_if_fail=False)

    Seed.set_seed(
        0, False
    )  # previous test (in the pipeline) changed the deterministic behavior to True

    run_train(paths, train_common_params)
    run_infer(paths, infer_common_params)
    results = run_eval(paths, eval_common_params)

    assert "metrics.auc" in results


class ClassificationISICTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.root = tempfile.mkdtemp()
        self.train_common_params = TRAIN_COMMON_PARAMS
        self.train_common_params["trainer.num_epochs"] = 1
        self.train_common_params["data.samples_ids"] = FULL_GOLDEN_MEMBERS

        self.infer_common_params = INFER_COMMON_PARAMS
        self.eval_common_params = EVAL_COMMON_PARAMS

    def test_runner_transformer(self) -> None:
        # Define model parameters
        token_dim = 768
        self.train_common_params["model_type"] = "Transformer"
        self.train_common_params["model"] = dict(
            token_dim=token_dim,
            projection_kwargs=dict(
                image_shape=[300, 300], patch_shape=[30, 30], channels=3
            ),
            transformer_kwargs=dict(
                depth=12,
                heads=12,
                mlp_dim=token_dim * 4,
                dim_head=64,
                dropout=0.0,
                emb_dropout=0.0,
            ),
        )

        # Reflect model choice in the infer stage
        self.infer_common_params["model_type"] = self.train_common_params["model_type"]
        self.infer_common_params["model"] = self.train_common_params["model"]
        run_in_subprocess(
            run_isic,
            self.root,
            self.train_common_params,
            self.infer_common_params,
            self.eval_common_params,
        )

    def test_runner_resnet18(self) -> None:
        # Define model parameters for two CNN backbone types
        self.train_common_params["model_type"] = "CNN"
        self.train_common_params["model"] = dict(
            dropout_rate=0.5,
            layers_description=(256,),
            tabular_data_inputs=[("data.input.clinical.all", 19)],
            tabular_layers_description=(128,),
            backbone_type="Resnet18",
        )

        # Reflect model choice in the infer stage
        self.infer_common_params["model_type"] = self.train_common_params["model_type"]
        self.infer_common_params["model"] = self.train_common_params["model"]
        run_in_subprocess(
            run_isic,
            self.root,
            self.train_common_params,
            self.infer_common_params,
            self.eval_common_params,
        )

    def test_runner_inception(self) -> None:
        # Define model parameters for two CNN backbone types
        self.train_common_params["model_type"] = "CNN"
        self.train_common_params["model"] = dict(
            dropout_rate=0.5,
            layers_description=(256,),
            tabular_data_inputs=[("data.input.clinical.all", 19)],
            tabular_layers_description=(128,),
            backbone_type="InceptionResnetV2",
        )

        # Reflect model choice in the infer stage
        self.infer_common_params["model_type"] = self.train_common_params["model_type"]
        self.infer_common_params["model"] = self.train_common_params["model"]
        run_in_subprocess(
            run_isic,
            self.root,
            self.train_common_params,
            self.infer_common_params,
            self.eval_common_params,
        )

    def tearDown(self) -> None:
        # Delete temporary directories
        shutil.rmtree(self.root)


if __name__ == "__main__":
    unittest.main()



import torch

from fuse.utils import NDict


class LossExtractFromBatchDict(torch.nn.Module):
    """
    Read the loss from batch dict
    Useful in cases that the model is the one that set the loss
    """

    def __init__(self, key: str, weight: float = 1.0) -> None:
        """
        :param key: batch_dict key to read the loss from
        :param weight: loss weight, the returned value will be multiplication of loss and weight
        """
        super().__init__()
        self.key = key
        self.weight = weight

    def forward(self, batch_dict: NDict) -> torch.Tensor:
        return batch_dict[self.key] * self.weight



special_token_marker = [
    "<",
    ">",
]


def special_wrap_input(x: str) -> str:
    return special_token_marker[0] + x + special_token_marker[1]


def strip_special_wrap(x: str) -> str:
    for spec_wrap in special_token_marker:
        x = x.replace(spec_wrap, "")
    return x



from fuse.utils import NDict
from fuse.data import OpBase, get_sample_id
from fuse.data.tokenizers.modular_tokenizer.modular_tokenizer import ModularTokenizer
from fuse.data.tokenizers.modular_tokenizer.inject_utils import (
    InjectorToModularTokenizerLib,
)
from huggingface_hub import snapshot_download, HfApi
from huggingface_hub.utils import validate_hf_hub_args, SoftTemporaryDirectory


from warnings import warn
from pathlib import Path
from collections import defaultdict
from typing import Any, Tuple, Dict, List, Optional, Union
import os
import re


class ModularTokenizerWithoutInjectOp(OpBase):
    """
    Modular tokenizer without an option to inject into the model scalars or embedding
    """

    def __init__(
        self,
        tokenizer_path: str,
        max_size: Union[int, None] = None,
        pad_token: Union[str, None] = None,
        pad_type_id: Union[int, None] = None,
        validate_ends_with_eos: Optional[bool] = True,
        eos: Optional[str] = "<EOS>",
        verbose: Optional[bool] = False,
        on_unknown_default_value: str = "warn",
        **kwargs: Any,
    ) -> None:
        """
        Args:
            tokenizer_path: full path to a directory that the tokenizer will be loaded from
            max_size: sequences below this size will be padded, and above this size will be truncated
            pad: a string of the pad token
            pad_type_id: see tokenizers.Tokenizer.enable_padding() docstring
            validate_ends_with_eos: during encoder request (a _call_ to the op) will make sure that it ends with the provided eos token, and raise exception otherwise.
                having an eos (end of sentence) token in the end is useful for multiple scenarios, for example in a generative transformer (like T5 encoder-decoder)
            verbose:
            on_unknown_default_value: User can define the default behavior of unknown token here in the constructor. In addition, this value can be overwritten in the __call__
        """
        super().__init__(**kwargs)

        if verbose:
            print(
                f"DEBUG:{self.__class__.__name__} __init__ called for path {tokenizer_path}"
            )

        self._tokenizer_path = tokenizer_path
        self._tokenizer = ModularTokenizer.from_file(self._tokenizer_path)
        pad_id = self._tokenizer.token_to_id(pad_token)

        if pad_token is None:
            raise Exception(
                f"Could not find pad token = {pad_token} in {tokenizer_path}"
            )

        self._validate_ends_with_eos = validate_ends_with_eos
        self._eos = eos
        self._on_unknown_default_value = on_unknown_default_value

        if on_unknown_default_value not in ["warn", "raise"]:
            raise ValueError(f"Doesn't support {on_unknown_default_value=}!")

        if self._validate_ends_with_eos:
            eos_id = self._tokenizer.token_to_id(self._eos)
            if eos_id is None:
                raise Exception(
                    f"Could not find eos token = {self._eos} in {tokenizer_path}. You can disable the validation by setting validate_ends_with_eos=False"
                )

        self._pad_id = pad_id
        self._verbose = verbose

        if max_size is not None:
            assert isinstance(max_size, int)
            assert max_size > 0
            assert isinstance(pad_id, int)

            padding_kwargs = dict(length=max_size, pad_id=pad_id)
            if pad_type_id is not None:
                assert isinstance(pad_type_id, int)
                padding_kwargs["pad_type_id"] = pad_type_id

            self._tokenizer.enable_padding(direction="right", **padding_kwargs)

            self._tokenizer.enable_truncation(
                max_length=max_size,
                direction="right",
            )

        self._max_size = max_size

        if self._verbose:
            self._debug_max_tokenized_len_encountered = defaultdict(int)

    def get_vocab_size(self) -> int:
        return self._tokenizer.get_vocab_size()

    def get_max_token_id(self) -> Tuple[str, int]:
        """
        scans the vocab for the max observed token id and returns a tuple for it
            [its token string (str), the token id (int)]
        """
        max_token_id = self._tokenizer.get_max_id()
        token_str_of_max_token_id = self._tokenizer.id_to_token(max_token_id)

        if max_token_id is None:
            raise Exception("Could not find max_token_id! possibly an empty vocab.")
        if token_str_of_max_token_id is None:
            warn(
                "max_token_id does not correspond to a token. It may be an upper bound placeholder."
            )
            return "placeholder upper-bound token", max_token_id
        return token_str_of_max_token_id, max_token_id

    def get_min_max_sentinels(
        self,
        sentinel_prefix: Optional[str] = "<SENTINEL_ID_",
        integer_find_regex: Optional[str] = r"\d{1,}",
    ) -> Tuple[int, int]:
        """
        returns a Tuple [min encountered sentinel name, max encountered sentinel name]

        For example, if the vocab contains:

        SENTINEL_ID_101: 1000,
        SENTINEL_ID_102: 1001,
        SENTINEL_ID_103: 1002,

        will return [101,103]
        """
        min_token = None
        max_token = None

        for k, _ in self._tokenizer.get_added_vocab().items():
            if sentinel_prefix in k:
                val = re.findall(integer_find_regex, k)
                if len(val) != 1:
                    raise Exception(
                        f"expected exactly one integer number in {k} but found {val}"
                    )
                val = val[0]
                val = int(val)

                if (min_token is None) or (val < min_token):
                    min_token = val

                if (max_token is None) or (val > max_token):
                    max_token = val

        if (min_token is None) or (max_token is None):
            raise Exception(
                f'Could not find any sentinels with the prefix "{sentinel_prefix}"'
            )

        return (min_token, max_token)

    def get_token_id(self, token_str: str, t_type: Optional[str] = None) -> int:
        """
        Args:
            token_str (:obj:`str`):
                The token to convert
            t_type (:obj:`str`): The sub-tokenizer to use. If None, the first (in order defined in the config)
                sub-tokenizer is used. If the token is special, type should not be set.

        Returns:
            :obj:`int`: The token's id under the tokenizer type (if given)
        """
        ans = self._tokenizer.token_to_id(token_str, t_type)
        assert ans is not None, f"could not find token id for token:{token_str}!"
        return ans

    def get_max_len(
        self, override_max_len: Union[int, None] = None
    ) -> Union[int, None]:
        """Returns the expected max_len of any encoding. max_len is given by internal state (set during initialization of the tokenizer), or it can be overridden
        during call to encode_list (applicable only to that specific encoding), or enable_padding/enable_truncation (applicable to all encodings produced
        following the call).

        Args:
            override_max_len (Optional[int], optional): Returns the resulting max_len, if the internal max_len were to be overridden by override_max_len
            during call to encode_list, or enable_padding/enable_truncation. Defaults to None.

        Returns:
            Optional[int]: _description_
        """
        return self._tokenizer.get_expected_max_len(override_max_len=override_max_len)

    def add_new_special_tokens(self, new_special_tokens: list[str]) -> int:
        """add new special tokens if they are not in the tokenizer.
        Skipps allready existing special tokens.

        Args:
            new_special_tokens (list[str]): the tokens to add
        Returns:
            `int`: The number of tokens that were created in the vocabulary

        Will raise an exception if any of the tokens are allready in the tokenizer as _regular_ tokens.
        """

        tokenizer = self._tokenizer
        num_new_tokens = tokenizer.add_special_tokens(new_special_tokens)
        return num_new_tokens

    def __call__(
        self,
        sample_dict: NDict,
        key_in: str,
        key_out_tokenized_object: Optional[str] = None,
        key_out_tokens_ids: Optional[str] = None,
        key_out_attention_mask: Optional[str] = None,
        convert_attention_mask_to_bool: Optional[bool] = True,
        max_seq_len: Optional[int] = None,
        on_unknown: Optional[str] = None,
        verbose: Optional[int] = 1,
        validate_ends_with_eos: Optional[bool] = None,
        additional_caller_info_text: Optional[str] = "",
        key_out_encoding_per_meta: Optional[str] = None,
    ) -> NDict:
        """_summary_

        Args:
            sample_dict (NDict): _description_
            key_in (str): key to either a:
                (1) string that contains, in addition to the text that is to be tokenized, special delimiters signifying the type
                of input within each span of text (e.g. <@TOKENIZER-TYPE=AA> sequence, <@TOKENIZER-TYPE=SMILES>, etc.).
                (2) list of modular_tokenizer.TypedInput specifying the tokenizer type and the subsequence to tokenize
            key_out_tokenized_object (Optional[str], optional): _description_. Defaults to None.
            key_out_tokens_ids (Optional[str], optional): _description_. Defaults to None.
            key_out_attention_mask (Optional[str], optional): _description_. Defaults to None.
            convert_attention_mask_to_bool (Optional[bool], optional): _description_. Defaults to True.
            max_seq_len (Optional[int], optional): set maximum sequence len dynamically, used for both padding and truncation.. Defaults to None.
            on_unknown (Optional[str], optional): What happens if unknown tokens (i.e. ones mapped to <UNK>) are encountered: 'raise' or 'warn'. Defaults to "warn". The default value can be determined in the constructor itself.
            verbose (Optional[int], optional): verbosity level. 0: no notification, 1: warning notification, 2: warning with partial data, 3: warning
                with full data. Defaults to 1.
            validate_ends_with_eos (Optional[bool], optional): if not None, overrides self._validate_ends_with_eos
            key_out_encoding_per_meta: optional key out. If set to a string will put in it the per-meta-instruction encoded parts as a list of Encoding elements

        Raises:
            Exception: _description_
            Exception: _description_

        Returns:
            NDict: _description_
        """
        data = sample_dict[key_in]
        if not isinstance(data, (list, str)):
            # data is a list of named tuples of type collections.namedtuple("TypedInput", ["input_type", "input_string", "max_len"])
            raise Exception(
                f"Expected key_in={key_in} to point to a list of inputs or string with builtin tokenizer hints, and instead got a {type(data)}. value={data}"
            )
        if validate_ends_with_eos is None:
            validate_ends_with_eos = self._validate_ends_with_eos

        if validate_ends_with_eos:
            if isinstance(data, str):
                last_seq = data
            else:
                last_seq = data[-1].input_string
            if not last_seq.rstrip().endswith(self._eos):
                raise Exception(
                    f"validate_ends_with_eos was set to {validate_ends_with_eos}, but about to encode a string that does not end with {self._eos}. The str end was: {last_seq}"
                )

        if on_unknown is None:
            # Use tokenizer instance default value
            on_unknown = self._on_unknown_default_value

        if isinstance(data, str):
            _ans = self._tokenizer.encode(
                data,
                max_len=max_seq_len,
                return_overflow_info=True,
                on_unknown=on_unknown,
                verbose=verbose,
                also_return_split=key_out_encoding_per_meta is not None,
            )
        else:
            _ans = self._tokenizer.encode_list(
                data,
                max_len=max_seq_len,
                return_overflow_info=True,
                on_unknown=on_unknown,
                verbose=verbose,
                also_return_split=key_out_encoding_per_meta is not None,
            )

        if key_out_encoding_per_meta is not None:
            encoded, overflow_info, per_meta_encoded = _ans
            sample_dict[key_out_encoding_per_meta] = per_meta_encoded
        else:
            encoded, overflow_info = _ans

        expected_max_len = self.get_max_len(override_max_len=max_seq_len)
        if (
            expected_max_len is not None
        ):  # we tightly couple padding length and max size.
            assert expected_max_len == len(encoded.ids)

        if self._verbose:
            if self._pad_id in encoded.ids:
                _encoded_len_unpadded = encoded.ids.index(self._pad_id)
            else:
                # no padding, therefore it was fully used (either exactly the size, or most likely it was clipped)
                _encoded_len_unpadded = len(encoded.ids)

            if (
                _encoded_len_unpadded
                > self._debug_max_tokenized_len_encountered[self._tokenizer_path]
            ):
                print(
                    f"DEBUG: {self.__class__.__name__} : encountered new max encoded size:",
                    _encoded_len_unpadded,
                    " for tokenizer: ",
                    self._tokenizer_path,
                )
                self._debug_max_tokenized_len_encountered[
                    self._tokenizer_path
                ] = _encoded_len_unpadded

        # KEEP THIS AS DOC FOR NOW
        # encoded has attributes [ids, type_ids, tokens, offsets, attention_mask, special_tokens_mask, overflowing]
        # ids are the encoded tokens,
        # type_ids are for things like "which sentence is this from"
        # tokens are the actual tokens (for example - ['c1ccc(', '/C(', '=N/N', 'c2nc3ccccc3', 's2)', 'c2cccc', 'n2)cc1', '[PAD]', '[PAD]', '[PAD]'])
        # offsets describe the starting point and length of each original token
        # attention_mask - by default puts 1 for everything that isn't padding, and 0 for those that are padding
        # special_tokens_mask - 1 for anything that is a special token (e.g. padding, separator, etc.) 0 for the rest
        # overflowing - it's encoded original content that get clipped out due to max length definition

        if (
            len(encoded.overflowing) > 0 and verbose > 0
        ):  # note, encoded.overflowing may have multiple items, and each item can contain multiple items
            print(
                f"Warning: {self.__class__.__name__}  (pid={os.getpid()}, {additional_caller_info_text}) had to truncate sequence: [{overflow_info}]  \
                    for tokenizer: {self._tokenizer_path} for sample_id {get_sample_id(sample_dict)}"
            )

        if key_out_tokenized_object is not None:
            # if requested, store the entire tokenizer.Encoding object (which provides access to attributes such as  [ids, type_ids, tokens, offsets, attention_mask, special_tokens_mask, overflowing])
            sample_dict[key_out_tokenized_object] = encoded

        if key_out_tokens_ids is not None:
            sample_dict[key_out_tokens_ids] = encoded.ids

        if key_out_attention_mask is not None:
            sample_dict[key_out_attention_mask] = encoded.attention_mask
            if convert_attention_mask_to_bool:
                sample_dict[key_out_attention_mask] = [
                    bool(x) for x in sample_dict[key_out_attention_mask]
                ]

        if (key_out_tokens_ids is None) and (key_out_tokenized_object is None):
            warn(
                f"{self.__class__.__name__}  Op got key_out_tokens_ids=None and key_out_tokenized_object=None, which means it will not modify anything in the sample. Is this intended?"
            )

        return sample_dict


# backward compatibility
FastModularTokenizer = ModularTokenizerWithoutInjectOp


class ModularTokenizerOp(ModularTokenizerWithoutInjectOp):
    """
    Extends ModularTokenizerWithoutInjectOp and adds the option to inject scalars or embedding as an input to the model.

    injector tokenizer builds on top of modular tokenizer op.
    its purpose is to build inputs_emb for the model (instead of input_ids)
        this allows to support more advanced inputs beyond token ids, like:
        * scalars inputs
        * embeddings vector within a single input

    supported syntax/format:

    for text following <@TOKENIZER-TYPE=SCALARS_LITERALS> supports the following format:
    ',' separated float values

    for text following <@TOKENIZER-TYPE=SCALARS_FROM_DICT> is expected to be a key to the sample NDict
        for example: "blah.boo.banana"  or "data.input.encoder_input"
        note: in SCALARS_FROM_DICT you can't describe masked scalars (outputs) you can only describe inputs

    example usage:

    encoder_input:
    <@TOKENIZER-TYPE=AA><MOLECULAR_WEIGHT_IN_SOME_UNIT><@TOKENIZER-TYPE=SCALARS_LITERALS>0.3<@TOKENIZER-TYPE=AA><BINDING_AFFINITY_NANOMOLAR><@TOKENIZER-TYPE=SCALARS_LITERALS><MASK><@TOKENIZER-TYPE=AA><SEQUENCE_NATURAL_START>ISGGDAIYSSTGRCSLGFNVRSGSTYYFLTAGICTDGATTWWANSARTTVLGTTSGSSFPNNDYGIVRYTNTTIPKDGTVGGQDITSAANATVGMAVTRRGSTTGTISGSVTALNATVNYGGGDVVYGMIRTNVCAEPGDSGGPLYSGTRAIGLTSGGSGNCSSGGTTFFQPVTEALVAYGVSVY<SEQUENCE_NATURAL_END>
    labels:
    <@TOKENIZER-TYPE=AA><MOLECULAR_WEIGHT_IN_SOME_UNIT><@TOKENIZER-TYPE=SCALARS_LITERALS>0.3<@TOKENIZER-TYPE=AA><BINDING_AFFINITY_NANOMOLAR><@TOKENIZER-TYPE=SCALARS_LITERALS>12.4<@TOKENIZER-TYPE=AA><SEQUENCE_NATURAL_START>ISGGDAIYSSTGRCSLGFNVRSGSTYYFLTAGICTDGATTWWANSARTTVLGTTSGSSFPNNDYGIVRYTNTTIPKDGTVGGQDITSAANATVGMAVTRRGSTTGTISGSVTALNATVNYGGGDVVYGMIRTNVCAEPGDSGGPLYSGTRAIGLTSGGSGNCSSGGTTFFQPVTEALVAYGVSVY<SEQUENCE_NATURAL_END>
    """

    def __init__(
        self,
        tokenizer_path: str,
        max_size: Union[int, None] = None,
        pad_token: Union[str, None] = None,
        pad_type_id: Union[int, None] = None,
        validate_ends_with_eos: Optional[bool] = True,
        eos: Optional[str] = "<EOS>",
        verbose: Optional[bool] = False,
        **kwargs: Any,
    ) -> None:
        """

        Args:
            tokenizer_path: full path to a directory that the tokenizer will be loaded from
            max_size: sequences below this size will be padded, and above this size will be truncated
            pad: a string of the pad token
            pad_type_id: see tokenizers.Tokenizer.enable_padding() docstring
            validate_ends_with_eos: during encoder request (a _call_ to the op) will make sure that it ends with the provided eos token, and raise exception otherwise.
                having an eos (end of sentence) token in the end is useful for multiple scenarios, for example in a generative transformer (like T5 encoder-decoder)
            verbose:
        """
        if verbose:
            print(
                f"DEBUG:{self.__class__.__name__} __init__ called for path {tokenizer_path}"
            )

        super().__init__(
            tokenizer_path=tokenizer_path,
            max_size=max_size,
            pad_token=pad_token,
            pad_type_id=pad_type_id,
            validate_ends_with_eos=validate_ends_with_eos,
            eos=eos,
            verbose=verbose,
            **kwargs,
        )

    def __call__(
        self,
        sample_dict: NDict,
        key_in: str,
        key_out_tokenized_object: Optional[str] = None,
        key_out_tokens_ids: Optional[str] = None,
        key_out_attention_mask: Optional[str] = None,
        convert_attention_mask_to_bool: Optional[bool] = True,
        max_seq_len: Optional[int] = None,
        on_unknown: Optional[str] = "warn",
        verbose: Optional[int] = 1,
        validate_ends_with_eos: Optional[bool] = None,
        key_out_scalars: Optional[str] = None,
    ) -> NDict:
        """_summary_

        Args:
            sample_dict (NDict): _description_
            key_in (str): key to either a:
                (1) string that contains, in addition to the text that is to be tokenized, special delimiters signifying the type
                of input within each span of text (e.g. <@TOKENIZER-TYPE=AA> sequence, <@TOKENIZER-TYPE=SMILES>, etc.).
                (2) list of modular_tokenizer.TypedInput specifying the tokenizer type and the subsequence to tokenize
            key_out_tokenized_object (Optional[str], optional): _description_. Defaults to None.
            key_out_tokens_ids (Optional[str], optional): _description_. Defaults to None.
            key_out_attention_mask (Optional[str], optional): _description_. Defaults to None.
            convert_attention_mask_to_bool (Optional[bool], optional): _description_. Defaults to True.
            max_seq_len (Optional[int], optional): set maximum sequence len dynamically, used for both padding and truncation.. Defaults to None.
            on_unknown (Optional[str], optional): What happens if unknown tokens (i.e. ones mapped to <UNK>) are encountered: 'raise' or 'warn'. Defaults to "warn".
            verbose (Optional[int], optional): verbosity level. 0: no notification, 1: warning notification, 2: warning with partial data, 3: warning
                with full data. Defaults to 1.
            validate_ends_with_eos (Optional[bool], optional): if not None, overrides self._validate_ends_with_eos
            key_out_scalars:str optional
                if provided, will write to:
                        `sample_dict[f'{key_out_scalars}.values]` - a 1D torch tensor with all the scalars values
                        `sample_dict[f'{key_out_scalars}.valid_mask]` - a 1D torch boolean tensor representing which elements have scalar values

        Returns:
            NDict: _description_
        """

        (
            with_placeholders_str,
            per_meta_orig,
        ) = InjectorToModularTokenizerLib.build_placeholder_meta_tokenization(
            sequence=sample_dict[key_in], sample_dict=sample_dict
        )
        sample_dict[key_in + ".with_placeholders"] = with_placeholders_str

        super().__call__(
            sample_dict=sample_dict,
            key_in=key_in + ".with_placeholders",
            key_out_tokenized_object=key_out_tokenized_object,
            key_out_tokens_ids=key_out_tokens_ids,
            key_out_attention_mask=key_out_attention_mask,
            convert_attention_mask_to_bool=convert_attention_mask_to_bool,
            max_seq_len=max_seq_len,
            on_unknown=on_unknown,
            verbose=verbose,
            validate_ends_with_eos=validate_ends_with_eos,
            key_out_encoding_per_meta=key_in
            + ".per_meta_part_encoding",  # using the key_in as base for the name because key_out_* are optional
        )

        prepared_data = InjectorToModularTokenizerLib.build_scalars(
            per_meta_tokenizer_data=per_meta_orig,
            per_meta_encoding_including_placeholders=sample_dict[
                key_in + ".per_meta_part_encoding"
            ],
            token_ids=sample_dict[key_out_tokens_ids],
            sample_dict=sample_dict,
        )

        if key_out_scalars is not None:
            sample_dict[key_out_scalars + ".values"] = prepared_data["scalars_values"]
            sample_dict[key_out_scalars + ".valid_mask"] = prepared_data[
                "scalars_valid_mask"
            ]

        return sample_dict

    @classmethod
    def from_pretrained(
        cls,
        identifier: str,
        pad_token: str = "<PAD>",
        max_size: Optional[int] = None,
        on_unknown_default_value: str = "warn",
        force_download: bool = False,
        resume_download: Optional[bool] = None,
        proxies: Optional[Dict] = None,
        token: Optional[Union[str, bool]] = None,
        cache_dir: Optional[Union[str, Path]] = None,
        local_files_only: bool = False,
        revision: Optional[str] = None,
    ) -> "ModularTokenizerOp":
        """Load pre-trained tokenizer from HF repo_id or a local dirpath.

        Args:
            identifier (str): A repo_id or local dirpath.
            pad_token (str, optional): A string of the pad token. Defaults to "<PAD>".
            max_size (Optional[int], optional): Sequences below this size will be padded, and above this size will be truncated. Defaults to None.
            * For other args see `snapshot_download()`
        """
        if not os.path.isdir(identifier):
            # Try to download from hub
            try:
                # Download 'tokenizer' folder from repo
                identifier = snapshot_download(
                    repo_id=str(identifier),
                    revision=revision,
                    cache_dir=cache_dir,
                    force_download=force_download,
                    proxies=proxies,
                    resume_download=resume_download,
                    token=token,
                    local_files_only=local_files_only,
                    allow_patterns="tokenizer/",
                )
                # Redirecting identifier to the downloaded folder
                identifier = os.path.join(identifier, "tokenizer")
            except Exception as e:
                raise Exception(
                    f"Couldn't find the checkpoint path nor download from HF hub! {identifier}"
                ) from e

        tokenizer_op = cls(
            tokenizer_path=identifier,
            pad_token=pad_token,
            max_size=max_size,
            on_unknown_default_value=on_unknown_default_value,
        )
        return tokenizer_op

    def save_pretrained(self, save_directory: Union[str, Path]) -> None:
        print(f"Saving @ {save_directory=}")
        self._tokenizer.save(path=str(save_directory))

    @validate_hf_hub_args
    def push_to_hub(
        self,
        repo_id: str,
        *,
        commit_message: str = "Push model using huggingface_hub.",
        private: bool = False,
        token: Optional[str] = None,
        branch: Optional[str] = None,
        create_pr: Optional[bool] = None,
        allow_patterns: Optional[Union[List[str], str]] = None,
        ignore_patterns: Optional[Union[List[str], str]] = None,
        delete_patterns: Optional[Union[List[str], str]] = None,
    ) -> None:
        api = HfApi(token=token)
        repo_id = api.create_repo(
            repo_id=repo_id, private=private, exist_ok=True
        ).repo_id
        # Push the files to the repo in a single commit
        with SoftTemporaryDirectory() as tmp:
            saved_path = Path(tmp) / repo_id
            tokenzier_dirpath = saved_path / "tokenizer"
            self.save_pretrained(tokenzier_dirpath)
            return api.upload_folder(
                repo_id=repo_id,
                repo_type="model",
                folder_path=saved_path,
                commit_message=commit_message,
                revision=branch,
                create_pr=create_pr,
                allow_patterns=allow_patterns,
                ignore_patterns=ignore_patterns,
                delete_patterns=delete_patterns,
            )



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Nov 30, 2023

"""

from fuse.eval.metrics.regression.metrics import MetricPearsonCorrelation
import numpy as np
import pandas as pd
from collections import OrderedDict
from fuse.eval.evaluator import EvaluatorDefault


def example_pearson_correlation() -> float:
    """
    Pearson correlation coefficient
    """

    # define data
    sz = 1000
    data = {
        "id": range(sz),
    }
    np.random.seed(0)
    rand_vec = np.random.randn((sz))
    data["x1"] = 100 * np.ones((sz)) + 10 * rand_vec
    data["x2"] = -10 * np.ones((sz)) + 3 * rand_vec

    data_df = pd.DataFrame(data)

    # list of metrics
    metrics = OrderedDict(
        [
            ("pearsonr", MetricPearsonCorrelation(pred="x1", target="x2")),
        ]
    )

    # read files
    evaluator = EvaluatorDefault()
    res = evaluator.eval(ids=None, data=data_df, metrics=metrics)

    return res



from typing import Callable, Optional

from fuse.dl.losses.loss_base import LossBase
from fuse.utils.ndict import NDict

import torch
import torch.nn.functional as F
from torch import Tensor


def softcrossentropyloss(target: Tensor, logits: Tensor) -> Tensor:
    """
    From the pytorch discussion Forum:
    https://discuss.pytorch.org/t/soft-cross-entropy-loss-tf-has-it-does-pytorch-have-it/69501
    """
    logprobs = F.log_softmax(logits, dim=1)
    loss = -(target * logprobs).sum() / target.sum()  # logits.shape[0]
    return loss


def supervised_contrastive_loss(
    feat_xi: Tensor,
    feat_xj: Tensor,
    yi: Tensor,
    yj: Tensor,
    symm: bool = False,
    alpha: float = 0.5,
    temp: float = 0.2,
) -> Tensor:
    """
    :param feat_i:               feature in the shape [Batch_size, embed dim]
    :param feat_j:               feature in the shape [Batch_size, embed dim]
    :param target_i:             ground truth label of class indexes in the shape [batch_size]
    :param target_j:             ground truth label of class indexes in the shape [batch_size]
    :param tempreture:           tempreture scaling parameter
    :param symmetrical:          should the output of the function is a combination of supervised contrastive loss components
    :param alpha:                if symmetrical -> the weight for each component
    """
    if feat_xi.isnan().any() | feat_xj.isnan().any():
        raise Exception("One of the features has nan value")

    feat_xi = F.normalize(feat_xi, p=2, dim=1)
    feat_xj = F.normalize(feat_xj, p=2, dim=1)
    label_vec = torch.unsqueeze(yi, 0)
    mask = torch.eq(torch.transpose(label_vec, 0, 1), yj).float()
    if mask.sum() == 0:
        return torch.tensor(0.0).to(mask.device)
    logits_xi = torch.div(torch.matmul(feat_xi, torch.transpose(feat_xj, 0, 1)), temp)
    logits_max, _ = torch.max(logits_xi, dim=1, keepdim=True)
    logits_xi = logits_xi - logits_max.detach()
    loss_xi = softcrossentropyloss(mask, logits_xi)  # / torch.sum(mask, 1)
    if symm:
        logits_xj = torch.matmul(feat_xj, torch.transpose(feat_xi, 0, 1)) / temp
        loss_xj = softcrossentropyloss(mask, logits_xj)
        return alpha * loss_xi.sum() + (1 - alpha) * loss_xj.sum()
    else:
        return loss_xi.mean()


class SupervisedContrastiveLoss(LossBase):
    """
    Supervised contrastive loss as defined in 'https://arxiv.org/pdf/2004.11362.pdf'

    """

    def __init__(
        self,
        *,  # prevent positional args
        feat_i: str = None,
        feat_j: str = None,
        target_i: str = None,
        target_j: str = None,
        tempreture: float = 0.2,
        alpha: float = 0.5,
        symmetrical: bool = False,
        callable: Callable = None,
        weight: Optional[float] = None,
        preprocess_func: Optional[Callable] = None,
    ) -> None:
        """
        This class wraps a PyTorch loss function with a Fuse api.
        Args:
        :param feat_i:               batch_dict key for feat i (e.g., network feature output)
        :param feat_j:               batch_dict key for feat_j (e.g., other network feature output to contrast with)
        :param target_i:             batch_dict key for target (e.g., ground truth label for feat_i)
        :param target_j:             batch_dict key for target (e.g., ground truth label for feat_j)
        :param tempreture:           tempreture scaling parameter
        :param symmetrical:          should the output of the function is a combination of supervised contrastive loss components
        :param alpha:                if symmetrical -> the weight for each component

        :param callable:           PyTorch loss function handle (e.g., torch.nn.functional.cross_entropy)
        :param weight:             scalar loss multiplier
        :param preprocess_func:             function that filters batch_dict/ The function gets an input batch_dict and returns filtered batch_dict
            the expected function signature is:
                foo(batch_dict: NDict) -> NDict:
        """
        super().__init__()
        self.feat_i = feat_i
        self.feat_j = feat_j
        self.target_i = target_i
        self.target_j = target_j

        self.temp = tempreture
        self.alpha = alpha
        self.symm = symmetrical

        self.callable = callable
        self.weight = weight
        self.preprocess_func = preprocess_func

    def forward(self, batch_dict: NDict) -> torch.Tensor:
        # preprocess batch_dict if required
        if self.preprocess_func is not None:
            batch_dict = self.preprocess_func(batch_dict)
        feat_i = batch_dict[self.feat_i]
        feat_j = batch_dict[self.feat_j]

        target_i = batch_dict[self.target_i]
        target_j = batch_dict[self.target_j]

        loss_obj = supervised_contrastive_loss(
            feat_i,
            feat_j,
            target_i,
            target_j,
            symm=self.symm,
            alpha=self.alpha,
            temp=self.temp,
        )
        if self.weight is not None:
            loss_obj *= self.weight

        return loss_obj



from typing import Sequence, Union, Dict, Optional
import numpy as np
import matplotlib.pyplot as plt
import math
import torch
from torch import nn, optim, Tensor
from scipy import stats


class Calibration:
    """
    Methods for probability calibration
    """

    @staticmethod
    def reliability_diagram(
        pred: Sequence[Union[np.ndarray, float]],
        target: Sequence[Union[np.ndarray, int]],
        num_bins: Optional[int] = 10,
        num_quantiles: Optional[int] = None,
        output_filename: Optional[str] = None,
    ) -> Dict:
        """
        Reliability diagram
        :param pred: List of arrays of shape [NUM_CLASSES]
        :param target: List of integers [0, NUM_CLASSES) denoting ground truth classes
        :num_bins: Integer denoting the number of equal width bins in the diagram's x-axis.
        :num_quantiles: Integer denoting the number of equal number of samples quantiles in the diagram's x-axis.
            Note, either num_bins or num_quantiles should be set to None, as they define different ways to split the x-axis.
        :output_filename: Path to file to which to save the plot
        :return: saving roc curve to a file and return the input to figure to a dictionary
        """
        assert num_bins is None or num_quantiles is None

        if isinstance(pred[0], float):  # binary case
            pred = [np.array([1 - p, p]) for p in pred]

        max_pred = [np.max(p) for p in pred]
        cls_pred = [np.argmax(p) for p in pred]

        assert all(
            [math.isclose(p.sum(), 1, rel_tol=1e-5) for p in pred]
        )  # make sure that the inputs make proper probabilities

        total_samples = len(pred)
        max_pred = np.array(max_pred)
        if num_bins is not None:
            conf_vec = np.linspace(0, 1, num_bins + 1)
        else:
            quantiles_vec = np.linspace(0, 1, num_quantiles + 1)
            conf_vec = stats.mstats.mquantiles(max_pred, quantiles_vec.tolist())
        acc_vec = np.zeros(len(conf_vec) - 1)
        fraction_of_samples = np.zeros(len(conf_vec) - 1)
        num_samples = np.zeros(len(conf_vec) - 1)
        for i in range(len(conf_vec) - 1):
            cnt_correct_in_conf = sum(
                [
                    p_cls == t
                    for (p_cls, p, t) in zip(cls_pred, max_pred, target)
                    if p >= conf_vec[i] and p < conf_vec[i + 1]
                ]
            )
            cnt_conf = len(
                [p for p in max_pred if p >= conf_vec[i] and p < conf_vec[i + 1]]
            )
            num_samples[i] = cnt_conf
            fraction_of_samples[i] = cnt_conf / np.max(
                [total_samples, np.finfo(float).eps]
            )
            acc_vec[i] = cnt_correct_in_conf / np.max([cnt_conf, np.finfo(float).eps])
        mid_conf_vec = np.diff(conf_vec) / 2 + conf_vec[:-1]
        cnt_correct_total = sum([p_cls == t for (p_cls, t) in zip(cls_pred, target)])
        avg_acc = cnt_correct_total / total_samples
        avg_conf = np.mean(max_pred)
        # extract info for the plot
        results = {}
        results["conf_vec"] = conf_vec
        results["mid_confidence_points"] = mid_conf_vec
        results["accuracy"] = acc_vec
        results["avg_accuracy"] = avg_acc
        results["avg_confidence"] = avg_conf
        results["num_samples_per_bin"] = num_samples
        results["total_samples"] = total_samples

        # display
        if output_filename is not None:
            f, (ax1, ax2) = plt.subplots(2, 1, sharex=True, sharey=True)
            f.tight_layout(pad=3.0)
            ax1.plot(
                mid_conf_vec, fraction_of_samples, color="blue", marker="o", linewidth=2
            )
            ax1.set_title(
                f"Confidence Histogram. Avg. Accuracy: {np.round(avg_acc, 2)}, Avg. Confidence: {np.round(avg_conf, 2)}"
            )
            ax1.set_xlabel("Confidence")
            ax1.set_ylabel("Fraction of samples")
            ax1.grid()
            ax2.plot(conf_vec, conf_vec, color="black", linewidth=2)
            ax2.plot(mid_conf_vec, acc_vec, color="red", marker="o", linewidth=2)
            ax2.set_title("Reliability Curve")
            ax2.set_xlabel("Confidence")
            ax2.set_ylabel("Accuracy")
            ax2.grid()
            f.savefig(output_filename)
            plt.close()

        return results

    @staticmethod
    def ece(
        pred: Sequence[Union[np.ndarray, float]],
        target: Sequence[Union[np.ndarray, int]],
        num_bins: Optional[int] = 10,
        num_quantiles: Optional[int] = None,
    ) -> float:
        """
        :param pred: prediction array per sample. Each element shape [num_classes]
        :param target: target per sample. Each element is an integer in range [0 - num_classes)
        :num_bins: Integer denoting the number of equal width bins in the diagram's x-axis.
        :num_quantiles: Integer denoting the number of equal number of samples quantiles in the diagram's x-axis.
            Note, either num_bins or num_quantiles should be set to None, as they define different ways to split the x-axis.
        :return Expected Calibration Error (ECE) score
        """
        reliability_results = Calibration.reliability_diagram(
            pred,
            target,
            num_bins=num_bins,
            num_quantiles=num_quantiles,
            output_filename=None,
        )
        conf_vec = reliability_results["conf_vec"][1:]
        total_samples = reliability_results["total_samples"]
        num_samples_per_bin = reliability_results["num_samples_per_bin"]
        existing_bins = num_samples_per_bin > 0
        accuracy_vec = reliability_results["accuracy"]
        # expected calibration error
        ece = (1.0 / total_samples) * np.sum(
            num_samples_per_bin * np.abs(accuracy_vec - conf_vec)
        )
        # maximum calibration error
        # we filter out bins which don't contain any samples
        mce = np.max(np.abs(accuracy_vec[existing_bins] - conf_vec[existing_bins]))

        results = {}
        results["ece"] = ece
        results["mce"] = mce

        return results

    @staticmethod
    def find_temperature(
        pred: Sequence[Union[np.ndarray, float]],
        target: Sequence[Union[np.ndarray, int]],
    ) -> float:
        """
        :param pred: prediction array per sample. Each element shape [num_classes]
            note that pred should be logits *before* conversion to softmax probabilities
        :param target: target per sample. Each element is an integer in range [0 - num_classes)
        :return temperature scale for calibration
        We optimize to find the temperature similarly to: https://github.com/gpleiss/temperature_scaling
        """
        nll_criterion = nn.CrossEntropyLoss()
        temperature = nn.Parameter(torch.ones(1) * 1.5)
        optimizer = optim.LBFGS([temperature], lr=0.01, max_iter=50)
        with torch.no_grad():
            logits = torch.tensor(np.array(pred))
            target = torch.tensor(target)

        def eval() -> Tensor:
            optimizer.zero_grad()
            loss = nll_criterion(logits / temperature, target)
            loss.backward()
            return loss

        optimizer.step(eval)

        return temperature.item()

    @staticmethod
    def apply_temperature(
        pred: Sequence[Union[np.ndarray, float]], temperature: Union[float, None] = None
    ) -> float:
        """
        :param pred: prediction array per sample. Each element shape [num_classes]
            note that pred should be logits *before* conversion to softmax probabilities
        :param temperature: temperature scalar
        :return array of softmax probabilities
        """
        if temperature is None:
            pred_calibrated = pred
        else:
            if isinstance(pred[0], np.ndarray):
                pred = np.array(pred)
            pred_calibrated = torch.tensor(pred / temperature)
            pred_calibrated = nn.Softmax(dim=1)(pred_calibrated)

        return np.array(pred_calibrated)



from lifelines.utils import concordance_index
import numpy as np
from typing import Sequence


class MetricsSurvival:
    @staticmethod
    def c_index(
        pred: np.ndarray,
        event_times: np.ndarray,
        event_observed: np.ndarray,
        event_class_index: int = -1,
        time_unit: int = 1,
        time_followup: int = None,
    ) -> float:
        """
        Compute c-index (concordance index) score using lifelines
        :param pred: prediction array per sample. Each element is a score (scalar). Higher score - higher chance for the event
        :param event_times: event/censor time array
        :param event_observed: Optioal- a length-n iterable censoring flags, 1 if observed, 0 if not.
        :param event_class_index: Optional - the index of the event class in predicted scores tuples
        :return c-index (concordance index) score
        """

        if isinstance(pred, Sequence):
            pred = np.asarray(pred)
        if isinstance(pred[0], np.ndarray):
            pred = pred[:, event_class_index]

        event_times = (np.array(event_times) / time_unit).astype(int)

        if time_followup is not None:
            event_observed = np.array(event_observed)
            event_times = np.array(event_times)
            print(
                f"C-index time_follow_up={time_followup}: ignored events ={event_observed[event_times>time_followup].sum()} remaining_events={event_observed[event_times<=time_followup].sum()}"
            )
            event_observed[event_times > time_followup] = 0

        return concordance_index(event_times, -pred, event_observed)

    @staticmethod
    def expected_cindex(
        pred: np.ndarray,
        event_times: np.ndarray,
        event_observed: np.ndarray,
    ) -> float:
        """
        Compute expected c-index (concordance index) score when given a survival distribution
        :param pred: a survival distribution per sample
        :param event_times: event/censor time array
        :param event_observed: a length-n iterable censoring flags, 1 if observed, 0 if not.
        :return expected c-index (concordance index) score
        """
        pred = np.array(pred)
        event_times = np.array(event_times)
        event_observed = np.array(event_observed)
        # calculating the "expected time" until an event
        num_bins = pred.shape[1]
        # [1,...,0] - highest risk
        # [0,...,1] - lowest risk
        expected_event_time = (pred * np.arange(num_bins)).sum(axis=1)
        return concordance_index(event_times, expected_event_time, event_observed)



"""
(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, Hashable, List, Optional, Sequence, Union, Any

from warnings import warn
from fuse.data.datasets.dataset_base import DatasetBase
from fuse.data.ops.ops_common import OpCollectMarker
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.utils.ndict import NDict
from fuse.utils.multiprocessing.run_multiprocessed import (
    run_multiprocessed,
    get_from_global_storage,
)
from fuse.data import (
    get_sample_id,
    create_initial_sample,
    get_specific_sample_from_potentially_morphed,
)
import copy
from collections import OrderedDict
import numpy as np
from operator import itemgetter


class DatasetDefault(DatasetBase):
    def __init__(
        self,
        sample_ids: Union[int, Sequence[Hashable], None],
        static_pipeline: Optional[PipelineDefault] = None,
        dynamic_pipeline: Optional[PipelineDefault] = None,
        cacher: Optional[SamplesCacher] = None,
        allow_uncached_sample_morphing: bool = False,
    ):
        """
        :param sample_ids: list of sample_ids included in dataset. Or:
                - An integer that describes only the size of the dataset. This is useful in massive datasets
                    (for example 100M samples). In such case, multiple functionalities will not be supported, mainly -
                    cacher, allow_uncached_sample_morphing and get_all_sample_ids
                - None. In this case, the dataset will not deal with sample ids. it is the user's responsibility to handle
                    iterations w.r.t the length of the dataset, as well as the index passed to __getitem__
                    this is useful for massive datasets, but when the sample ids are not expected to be running integets from 0 to a given length.

        :param static_pipeline:   static_pipeline, the output of this pipeline will be automatically cached.
        :param dynamic_pipeline:  dynamic_pipeline. applied sequentially after the static_pipeline, but not automatically cached.
                                changing it will NOT trigger recaching of the static_pipeline part.
        :param cacher: optional SamplesCacher instance which will be used for caching samples to speed up samples loading
        :param allow_uncached_sample_morphing:  when enabled, allows an Op, to return None, or to return multiple samples (in a list)

        """
        super().__init__()

        # store arguments
        self._cacher = cacher
        if isinstance(sample_ids, (int, np.integer)):
            if allow_uncached_sample_morphing:
                raise Exception(
                    "allow_uncached_sample_morphing is not allowed when providing sample_ids=an integer value"
                )
            if cacher is not None:
                raise Exception(
                    "providing a cacher is not allowed when providing sample_ids=an integer value"
                )
            self._sample_ids_mode = "running_int"
        elif sample_ids is None:
            self._sample_ids_mode = "external"
        else:
            self._sample_ids_mode = "explicit"

        # self._orig_sample_ids = sample_ids
        self._allow_uncached_sample_morphing = allow_uncached_sample_morphing

        # verify unique names for dynamic pipelines
        if dynamic_pipeline is not None and static_pipeline is not None:
            if static_pipeline.get_name() == dynamic_pipeline.get_name():
                raise Exception(
                    f"Detected identical name for static pipeline and dynamic pipeline ({static_pipeline.get_name(static_pipeline.get_name())}).\nThis is not allowed, please initiate the pipelines with different names."
                )

        if static_pipeline is None:
            static_pipeline = PipelineDefault(
                "dummy_static_pipeline", ops_and_kwargs=[]
            )
        if dynamic_pipeline is None:
            dynamic_pipeline = PipelineDefault(
                "dummy_dynamic_pipeline", ops_and_kwargs=[]
            )

        if dynamic_pipeline is not None:
            assert isinstance(
                dynamic_pipeline, PipelineDefault
            ), f"dynamic_pipeline may be None or a PipelineDefault instance. Instead got {type(dynamic_pipeline)}"

        if static_pipeline is not None:
            assert isinstance(
                static_pipeline, PipelineDefault
            ), f"static_pipeline may be None or a PipelineDefault instance. Instead got {type(static_pipeline)}"

        if self._allow_uncached_sample_morphing:
            warn(
                "allow_uncached_sample_morphing is enabled! It is a significantly slower mode and should be used ONLY for debugging"
            )

        self._static_pipeline = static_pipeline
        self._dynamic_pipeline = dynamic_pipeline
        self._orig_sample_ids = copy.deepcopy(sample_ids)

        self._created = False

    @property
    def static_pipeline(self) -> Union[PipelineDefault, None]:
        return self._static_pipeline

    @property
    def dynamic_pipeline(self) -> Union[PipelineDefault, None]:
        return self._dynamic_pipeline

    def create(self, num_workers: int = 0, mp_context: Optional[str] = None) -> None:
        """
        Create the data set, including caching
        :param num_workers: number of workers. used only when caching is disabled and allow_uncached_sample_morphing is enabled
            set num_workers=0 to disable multiprocessing (more convenient for debugging)
            Setting num_workers for caching is done in cacher constructor.
        :param mp_context: "fork", "spawn", "thread" or None for multiprocessing default
        :return: None
        """

        self._output_sample_ids_info = None
        if self._cacher is not None:
            self._output_sample_ids_info = self._cacher.cache_samples(
                self._orig_sample_ids
            )
        elif self._allow_uncached_sample_morphing:
            _output_sample_ids_info_list = run_multiprocessed(
                DatasetDefault._process_orig_sample_id,
                [(sid, self._static_pipeline, False) for sid in self._orig_sample_ids],
                workers=num_workers,
                mp_context=mp_context,
                desc="dataset_default.sample_morphing",
            )

            self._output_sample_ids_info = OrderedDict()
            self._final_sid_to_orig_sid = {}
            for sample_in_out_info in _output_sample_ids_info_list:
                orig_sid, out_sids = sample_in_out_info[0], sample_in_out_info[1]
                self._output_sample_ids_info[orig_sid] = out_sids
                if out_sids is not None:
                    assert isinstance(out_sids, list)
                    for final_sid in out_sids:
                        self._final_sid_to_orig_sid[final_sid] = orig_sid

        if self._output_sample_ids_info is not None:  # sample morphing is allowed
            self._final_sample_ids = []
            for orig_sid, out_sids in self._output_sample_ids_info.items():
                if out_sids is None:
                    continue
                self._final_sample_ids.extend(out_sids)
        else:
            self._final_sample_ids = self._orig_sample_ids

        self._orig_sample_ids = (
            None  # should not be use after create. use self._final_sample_ids instead
        )
        self._created = True

    def get_all_sample_ids(self) -> List[Any]:
        if not self._created:
            raise Exception("you must first call create()")

        if self._sample_ids_mode != "explicit":
            raise Exception(
                "get_all_sample_ids is not supported when constructed with non explicit sample_ids"
            )

        return copy.deepcopy(self._final_sample_ids)

    def set_final_sample_ids(self, final_sample_ids: List[Any]) -> None:
        self._final_sample_ids = final_sample_ids

    def add_sample_ids_and_recreate(
        self, added_sample_ids: List[Any], **kwargs: Dict
    ) -> None:
        self._orig_sample_ids = self._final_sample_ids + added_sample_ids
        self._created = False
        self.create(**kwargs)

    def __getitem__(self, item: Union[int, Hashable]) -> NDict:
        """
        Get sample, read from cache if possible
        :param item: either int representing sample index or sample_id
        :return: sample_dict
        """
        return self.getitem(item)

    def getitem(
        self,
        item: Union[int, Hashable],
        collect_marker_name: Optional[str] = None,
        keys: Optional[Sequence[str]] = None,
    ) -> NDict:
        """
        Get sample, read from cache if possible
        :param item: either int representing sample index or sample_id
        :param collect_marker_name: Optional, specify name of collect marker op to optimize the running time
        :param keys: Optional, return just the specified keys or everything available if set to None
        :return: sample_dict
        """
        if not self._created:
            raise Exception("you must first call create()")

        # get sample id
        if self._sample_ids_mode != "explicit":
            sample_id = item
            if self._sample_ids_mode == "running_int":  # allow using non int sample_ids
                if sample_id >= self._final_sample_ids:
                    raise IndexError(
                        f"Expecting {sample_id} to be smaller than {self._final_sample_ids}"
                    )

        elif not isinstance(item, (int, np.integer)):
            sample_id = item
        else:
            sample_id = self._final_sample_ids[item]

        # get collect marker info
        collect_marker_info = self._get_collect_marker_info(collect_marker_name)

        # read sample
        if self._cacher is not None:
            sample = self._cacher.load_sample(
                sample_id, collect_marker_info["static_keys_deps"]
            )

        if self._cacher is None:
            if not self._allow_uncached_sample_morphing:
                sample = create_initial_sample(sample_id)
                sample = self._static_pipeline(sample)
                if not isinstance(sample, dict):
                    raise Exception(
                        f'By default when caching is disabled sample morphing is not allowed, and the output of the static pipeline is expected to be a dict. Instead got {type(sample)}. You can use "allow_uncached_sample_morphing=True" to allow this, but be aware it is slow and should be used only for debugging'
                    )
            else:
                orig_sid = self._final_sid_to_orig_sid[sample_id]
                sample = create_initial_sample(orig_sid)
                sample = self._static_pipeline(sample)

                assert sample is not None
                sample = get_specific_sample_from_potentially_morphed(sample, sample_id)

        sample = self._dynamic_pipeline(
            sample, until_op_id=collect_marker_info["op_id"]
        )

        if not isinstance(sample, dict):
            raise Exception(
                f"The final output of dataset static (+optional dynamic) pipelines is expected to be a dict. Instead got {type(sample)}"
            )

        # get just required keys
        if keys is not None:
            sample = sample.get_multi(keys)

        return sample

    def _get_multi_multiprocess_func(self, args: Any) -> Any:
        sid, kwargs = args
        return self.getitem(sid, **kwargs)

    @staticmethod
    def _getitem_multiprocess(item: Union[Hashable, int, np.integer]) -> Any:
        """
        getitem method used to optimize the running time in a multiprocess mode
        """
        dataset = get_from_global_storage("dataset_default_get_multi_dataset")
        kwargs = get_from_global_storage("dataset_default_get_multi_kwargs")
        return dataset.getitem(item, **kwargs)

    def get_multi(
        self,
        items: Optional[Sequence[Union[int, Hashable]]] = None,
        workers: int = 10,
        verbose: int = 1,
        mp_context: Optional[str] = None,
        desc: str = "dataset_default.get_multi",
        **kwargs: Any,
    ) -> List[Dict]:
        """
        See super class
        :param workers: number of processes to read the data. set to 0 to not use multi processing (useful when debugging).
        :param mp_context: "fork", "spawn", "thread" or None for multiprocessing default
        """
        if items is None:
            sample_ids = list(range(len(self)))
        else:
            sample_ids = items

        for_global_storage = {
            "dataset_default_get_multi_dataset": self,
            "dataset_default_get_multi_kwargs": kwargs,
        }

        list_sample_dict = run_multiprocessed(
            worker_func=self._getitem_multiprocess,
            copy_to_global_storage=for_global_storage,
            args_list=sample_ids,
            workers=workers,
            verbose=verbose,
            mp_context=mp_context,
            desc=desc,
        )
        return list_sample_dict

    def __len__(self) -> int:
        if not self._created:
            raise Exception("you must first call create()")

        if self._sample_ids_mode == "running_int":
            return self._final_sample_ids
        elif self._sample_ids_mode == "external":
            raise Exception(
                "__len__ is not defined where explicit sample_ids or an interer len are not provided."
            )

        return len(self._final_sample_ids)

    # internal methods

    @staticmethod
    def _process_orig_sample_id(args: Any) -> Any:
        """
        Process, without caching, single sample
        """
        orig_sample_id, pipeline, return_sample_dict = args
        sample = create_initial_sample(orig_sample_id)

        sample = pipeline(sample)

        output_sample_ids = None

        if sample is not None:
            output_sample_ids = []
            if not isinstance(sample, list):
                sample = [sample]
            for curr_sample in sample:
                output_sample_ids.append(get_sample_id(curr_sample))

        if not return_sample_dict:
            return orig_sample_id, output_sample_ids

        return orig_sample_id, output_sample_ids, sample

    def _get_collect_marker_info(self, collect_marker_name: str) -> dict:
        """
        Find the required collect marker (OpCollectMarker in the dynamic pipeline).
        See OpCollectMarker for more details
        :param collect_marker_name: name to identify the required collect marker
        :return: a dictionary with the required info - including: name, op_id and static_keys_deps.
        if collect_marker_name is None will return default instruct to run the entire dynamic pipeline
        """
        # default values for case collect marker info is not used
        if collect_marker_name is None:
            return {"name": None, "op_id": None, "static_keys_deps": None}

        # find the required collect markers and extract the info
        collect_marker_info = None
        for (op, _), op_id in reversed(
            zip(self._dynamic_pipeline.ops_and_kwargs, self._dynamic_pipeline._op_ids)
        ):
            if isinstance(op, OpCollectMarker):
                collect_marker_info_cur = op.get_info()
                if collect_marker_info_cur["name"] == collect_marker_name:
                    if collect_marker_info is None:
                        collect_marker_info = collect_marker_info_cur
                        collect_marker_info["op_id"] = op_id
                        # continue to make sure this is the only one
                    else:
                        # throw an error if found more than one collect marker
                        raise Exception(
                            f"Error: two collect markers with name {collect_marker_info} found in dynamic pipeline"
                        )
        if collect_marker_info is None:
            raise Exception(
                f"Error: didn't find collect marker with name {collect_marker_info} in dynamic pipeline."
            )

        return collect_marker_info

    def summary(self) -> str:
        sum = ""
        sum += f"Type: {type(self).__name__}\n"
        sum += f"Num samples: {len(self._final_sample_ids)}\n"
        # TODO
        # sum += f"Cacher: {self._cacher.summary()}"
        # sum += f"Pipeline static: {self._static_pipeline.summary()}"
        # sum += f"Pipeline dynamic: {self._dynamic_pipeline.summary()}"

        return sum

    def subset(self, indices: Sequence[int]) -> None:
        """
        create a subset of the dataset by a given indices (inplace).

        Example:
            For the dataset '[-2, 1, 5, 3, 8, 5, 6]' and the indices '[1, 2, 5]', the subset is [1, 5, 5]

        :param items: indices of the subset - if None, the subset is the whole set.
        """
        if indices is None:
            # Do nothing, the subset is the whole dataset
            return

        if not self._created:
            raise Exception("you must first call create()")

        # grab the specified data
        self._final_sample_ids = itemgetter(*indices)(self._final_sample_ids)



from typing import List, Union
import torch
import functools


def select(
    condlist: List[torch.Tensor],
    choicelist: List[torch.Tensor],
    default_val: Union[float, int],
) -> torch.Tensor:
    """Mimicking numpy.select https://numpy.org/doc/stable/reference/generated/numpy.select.html
        by iteratively invoking np.where(condtion, choice, default_val) in reverse order
    :param List[torch.Tensor] condlist: The list of conditions which determine from which array in `choicelist`
        the output elements are taken. When multiple conditions are satisfied,
        the first one encountered in `condlist` is used.
    :param List[torch.Tensor] choicelist: The list of tensors from which the output elements are taken. It has
        to be of the same length as `condlist`.
    :param  Union[float, int] default_val: The element inserted in `output` when all conditions evaluate to False.
    :return torch.Tensor: The value at each position corresponds to the value in the corresponding position in the m-th choicelist
                          if the first 'True' value in the corresponding positions in condlist tensors is m.
                          When all corresponding positions in the condlist are 'False', the value will 'default_val'
    """
    args_list = reversed(list(zip(condlist, choicelist)))
    func = lambda d, args: torch.where(*args, other=d)
    return functools.reduce(func, args_list, default_val)



from typing import Dict, Optional, Sequence
import numpy as np
import scipy


class Ensembling:
    """
    Methods for ensembling
    """

    @staticmethod
    def ensemble(
        preds: Sequence[np.ndarray], method: Optional[str] = "average"
    ) -> Dict:
        """
        :param preds: sequence of numpy arrays / floats of shape [NUM_CLASSES]
        :params method: Ensembling method. 'average', or 'voting'
            if 'average', the predictions are assumed to be continuous (probabilities or regression output)
            if 'voting', the predictions are assumed to be class predictions (integers)
        """
        if isinstance(preds, Sequence):
            preds = np.stack(preds)

        # ensemble
        if method.lower() in ("average", "mean"):
            preds_ensembled = np.nanmean(preds, 1)
        elif method.lower() in ("vote", "voting"):
            assert len(preds.shape) == 2 or (
                len(preds.shape) == 3 and preds.shape[2] == 1
            )
            if len(preds.shape) == 3:
                preds = np.squeeze(preds)
            preds_ensembled = scipy.stats.mode(preds, axis=1)[0]
        else:
            raise ValueError(
                "'average' or 'voting' methods are supported for ensembling"
            )

        return preds_ensembled
